/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkRect2DActor.h,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkRect2DActor - Display a rectangle with border and background
// .SECTION Description
// vtkRect2DActor is a rectangle with a line border and a background. It is 
// normally used with vtkRect2DWidget to allow interactive rectangular
// selection in a render window. This initial position and size of the
// rectangle is the same as vtkActor2D.
//
// .SECTION See Also
// vtkRect2DWidget vtkActor2D

#ifndef __vtkRect2DActor_h
#define __vtkRect2DActor_h

#include "vtkActor2D.h"

class vtkPolyData;
class vtkPolyDataMapper2D;

class VTK_RENDERING_EXPORT vtkRect2DActor : public vtkActor2D
{
public:
  vtkTypeRevisionMacro(vtkRect2DActor,vtkActor2D);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Instantiate the rectangle with a border but no background. The default
  // position and size correspond to vtkActor2D.
  static vtkRect2DActor *New();

  // Description:
  // Draw the rectangle in the viewport.
  int RenderOpaqueGeometry(vtkViewport* viewport);
  int RenderTranslucentGeometry(vtkViewport*) { return 0; };
  int RenderOverlay(vtkViewport* viewport);

  // Description:
  // Release any graphics resources that are being consumed by this actor.
  // The parameter window could be used to determine which graphic
  // resources to release.
  virtual void ReleaseGraphicsResources(vtkWindow *);

  // Description:
  // Shallow copy of a rectangle actor. Overloads the virtual vtkProp method.
  void ShallowCopy(vtkProp *prop);

  // Description:
  // Turn the border on/off. The default border colour is black with line
  // width = 1.
  virtual void SetShowBorder( const int show );
  vtkGetMacro( ShowBorder, int );
  vtkBooleanMacro( ShowBorder, int );

  // Description:
  // Get the border actor.
  vtkGetObjectMacro( BorderActor, vtkActor2D );
  
  // Description:
  // Set/Get the border actor's vtkProperty2D.
  virtual void SetBorderProperty( vtkProperty2D* property2D );
  vtkGetObjectMacro( BorderProperty, vtkProperty2D );

  // Description:
  // Turn the rectangle background on/off. The default background color is white
  // with opacity = 0.3.
  virtual void SetShowBackground( const int show );
  vtkGetMacro( ShowBackground, int );
  vtkBooleanMacro( ShowBackground, int );

  // Description:
  // Get the background actor.
  vtkGetObjectMacro( BackgroundActor, vtkActor2D );

  // Description:
  // Set/Get the background actor's vtkProperty2D.
  virtual void SetBackgroundProperty( vtkProperty2D* property2D );
  vtkGetObjectMacro( BackgroundProperty, vtkProperty2D );

protected:
  vtkRect2DActor();
  ~vtkRect2DActor();

  vtkActor2D* BorderActor;
  vtkActor2D* BackgroundActor;

  void SizeRect2D( vtkViewport* viewport );

private:

  int                  ShowBorder;
  vtkPolyDataMapper2D* BorderMapper;
  vtkPolyData*         BorderPolyData;
  vtkProperty2D*       BorderProperty;
  
  int                  ShowBackground; 
  vtkPolyDataMapper2D* BackgroundMapper;
  vtkPolyData*         BackgroundPolyData;
  vtkProperty2D*       BackgroundProperty;

  int LastPosition[2];
  int LastPosition2[2];

  vtkTimeStamp BuildTime;

  vtkRect2DActor(const vtkRect2DActor&);  // Not implemented.
  void operator=(const vtkRect2DActor&);  // Not implemented.
};

#endif

