/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkRect2DWidget.cxx,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "stdafx.h"

#include "vtkRect2DWidget.h"
#include "vtkCallbackCommand.h"
#include "vtkObjectFactory.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"

vtkCxxRevisionMacro(vtkRect2DWidget, "$Revision: 1.0 $");
vtkStandardNewMacro(vtkRect2DWidget);

vtkRect2DWidget::vtkRect2DWidget()
{
  this->Rect2DActor = vtkRect2DActor::New();

  this->EventCallbackCommand->SetCallback(vtkRect2DWidget::ProcessEvents);
  this->State = vtkRect2DWidget::Outside;
  this->LeftButtonDown = 0;
  this->RightButtonDown = 0;
  this->Priority = 0.55;
}


vtkRect2DWidget::~vtkRect2DWidget()
{
  this->Rect2DActor->Delete();
}


void vtkRect2DWidget::PlaceRect( const double xMin, const double yMin,
                                 const double xMax, const double yMax )
{
   // The supplied position coordinates are NORMALISED VIEWPORT.
   double pos[2];
   pos[0] = xMin;
   pos[1] = yMin;
   this->Rect2DActor->SetPosition( pos );

   pos[0] = xMax - xMin;
   pos[1] = yMax - yMin;
   this->Rect2DActor->SetPosition2( pos );

   this->InvokeEvent( vtkCommand::PlaceWidgetEvent, NULL );  
}


void vtkRect2DWidget::SetEnabled( int enabling )
{
  if ( ! this->Interactor )
    {
    vtkErrorMacro( << "The interactor must be set prior to enabling/disabling widget" );
    return;
    }
  
  if ( enabling ) 
    {
    vtkDebugMacro( << "Enabling Rect2D widget" );
    if ( this->Enabled ) //already enabled, just return
      {
      return;
      }
    
    if ( ! this->CurrentRenderer )
      {
      this->SetCurrentRenderer(this->Interactor->FindPokedRenderer(
                               this->Interactor->GetLastEventPosition()[0],
                               this->Interactor->GetLastEventPosition()[1]) );
      if ( this->CurrentRenderer == NULL )
        {
        return;
        }
      }
    this->Enabled = 1;
    
    // Listen for the following events.
    vtkRenderWindowInteractor* i = this->Interactor;
    i->AddObserver( vtkCommand::MouseMoveEvent, 
                    this->EventCallbackCommand, this->Priority );
    i->AddObserver( vtkCommand::LeftButtonPressEvent, 
                    this->EventCallbackCommand, this->Priority );
    i->AddObserver( vtkCommand::LeftButtonReleaseEvent, 
                    this->EventCallbackCommand, this->Priority );
    i->AddObserver( vtkCommand::RightButtonPressEvent, 
                    this->EventCallbackCommand, this->Priority );
    i->AddObserver( vtkCommand::RightButtonReleaseEvent, 
                    this->EventCallbackCommand, this->Priority );

    // This is here to catch changes in the actor position/size due to resizing
    // of the render window. The event is sent by vtkRect2DActor::SizeRect2D().
    // The side effect is that 2 events are now sent by the widget for any
    // interaction event.
//    Rect2DActor->AddObserver( vtkCommand::UserEvent + 99, 
//                              this->EventCallbackCommand, this->Priority );
    Rect2DActor->AddObserver( vtkCommand::ModifiedEvent, 
                              this->EventCallbackCommand, this->Priority );

    // Add the rectangle actor.
    this->CurrentRenderer->AddProp( this->Rect2DActor );

    this->InvokeEvent( vtkCommand::EnableEvent, NULL );
    }
  else
    {
    vtkDebugMacro( << "Disabling Rect2D widget" );
    if ( ! this->Enabled ) //already disabled, just return
      {
      return;
      }
    this->Enabled = 0;

    // Don't listen for events any more.
    this->Interactor->RemoveObserver( this->EventCallbackCommand );

    // Remove the rectangle actors.
    this->CurrentRenderer->RemoveActor( this->Rect2DActor );

    this->InvokeEvent( vtkCommand::DisableEvent, NULL );
    this->SetCurrentRenderer( NULL );
    }
  
  this->Interactor->Render();
}


void vtkRect2DWidget::ProcessEvents(vtkObject* vtkNotUsed(object), 
                                       unsigned long event,
                                       void* clientdata, 
                                       void* vtkNotUsed(calldata))
{
  vtkRect2DWidget* self = reinterpret_cast<vtkRect2DWidget*>( clientdata );
  
  // Handle the mouse events.
  switch( event )
    {
    case vtkCommand::LeftButtonPressEvent:
      self->OnLeftButtonDown();
      break;
    case vtkCommand::LeftButtonReleaseEvent:
      self->OnLeftButtonUp();
      break;
    case vtkCommand::RightButtonPressEvent:
      self->OnRightButtonDown();
      break;
    case vtkCommand::RightButtonReleaseEvent:
      self->OnRightButtonUp();
      break;
    case vtkCommand::MouseMoveEvent:
      self->OnMouseMove();
      break;

//    case vtkCommand::UserEvent + 99:
    case vtkCommand::ModifiedEvent:
//      self->InvokeEvent(vtkCommand::InteractionEvent,NULL);
      self->InvokeEvent(vtkCommand::ModifiedEvent,NULL);
      break;

    }
}


int vtkRect2DWidget::ComputeStateBasedOnPosition( int X, int Y, int *pos1, int *pos2 )
{
  int result;
  
  // what are we modifying? The position, or size?
  // if size what piece?
  // if we are within 7 pixels of an edge...
  int e1 = 0;
  int e2 = 0;
  int e3 = 0;
  int e4 = 0;
  if (X - pos1[0] < 7)
    {
    e1 = 1;
    }
  if (pos2[0] - X < 7)
    {
    e3 = 1;
    }
  if (Y - pos1[1] < 7)
    {
    e2 = 1;
    }
  if (pos2[1] - Y < 7)
    {
    e4 = 1;
    }

  // assume we are moving
  result = vtkRect2DWidget::Moving;
  // unless we are on a corner or edges
  if (e2)
    {
    result = vtkRect2DWidget::AdjustingE2;
    }
  if (e4)
    {
    result = vtkRect2DWidget::AdjustingE4;
    }
  if (e1)
    {
    result = vtkRect2DWidget::AdjustingE1;
    if (e2)
      {
      result = vtkRect2DWidget::AdjustingP1;
      }
    if (e4)
      {
      result = vtkRect2DWidget::AdjustingP4;
      }
    }
  if (e3)
    {
    result = vtkRect2DWidget::AdjustingE3;
    if (e2)
      {
      result = vtkRect2DWidget::AdjustingP2;
      }
    if (e4)
      {
      result = vtkRect2DWidget::AdjustingP3;
      }
    }

  return result;
}


void vtkRect2DWidget::SetCursor(int cState)
{
  switch (cState)
    {
    case vtkRect2DWidget::AdjustingP1:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZESW);
      break;
    case vtkRect2DWidget::AdjustingP3:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZENE);
      break;
    case vtkRect2DWidget::AdjustingP2:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZESE);
      break;
    case vtkRect2DWidget::AdjustingP4:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZENW);
      break;
    case vtkRect2DWidget::AdjustingE1:
    case vtkRect2DWidget::AdjustingE3:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZEWE);
      break;
    case vtkRect2DWidget::AdjustingE2:
    case vtkRect2DWidget::AdjustingE4:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZENS);
      break;
    case vtkRect2DWidget::Moving:
      this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_SIZEALL);
      break;        
    }
}


void vtkRect2DWidget::OnLeftButtonDown()
{
  // This method can only be called when the widget is enabled.
  // Get the display coordinates (x,y pixel edges relative to the lower-left
  // corner of the render window). If the mouse is not confined to the render
  // window, these coordinates can be -ve.

  int X = this->Interactor->GetEventPosition()[0];
  int Y = this->Interactor->GetEventPosition()[1];

  // Get the display coordinates of the border actor defining the rectangle.
  int* pos1 = this->Rect2DActor->GetPositionCoordinate()
                               ->GetComputedDisplayValue(this->CurrentRenderer);
  int* pos2 = this->Rect2DActor->GetPosition2Coordinate()
                               ->GetComputedDisplayValue(this->CurrentRenderer);

  // Ignore event if the mouse is not over the rectangle.
  if ( X < pos1[0] || X > pos2[0] || Y < pos1[1] || Y > pos2[1] )
    {
    return;
    }
  
  // Start a drag. Convert the display coordinates of the mouse to normalised
  // viewport coordinates to match the 2D actor.

  double X2 = X;
  double Y2 = Y;
  this->CurrentRenderer->DisplayToNormalizedDisplay(X2,Y2);
  this->CurrentRenderer->NormalizedDisplayToViewport(X2,Y2);
  this->CurrentRenderer->ViewportToNormalizedViewport(X2,Y2);
  this->StartPosition[0] = X2;
  this->StartPosition[1] = Y2;

  this->State = this->ComputeStateBasedOnPosition(X, Y, pos1, pos2);
  this->SetCursor(this->State);
  
  // Do not pass this event onto any other interactor observers.
  this->EventCallbackCommand->SetAbortFlag(1);
  this->StartInteraction();
  this->InvokeEvent(vtkCommand::StartInteractionEvent,NULL);
  this->LeftButtonDown = 1;
}


void vtkRect2DWidget::OnMouseMove()
{
  // Get the display coordinates of the current mouse position (x,y pixel edges
  // relative to the lower-left corner of the render window). If the mouse is
  // not confined to the render window, these coordinates can be -ve.

  int X = this->Interactor->GetEventPosition()[0];
  int Y = this->Interactor->GetEventPosition()[1];

  // If the widget is not being dragged or resized (inside or outside), set the
  // inside/outside state.

  if ( this->State == vtkRect2DWidget::Outside ||
       this->State == vtkRect2DWidget::Inside )
    {
      int* pos1 = this->Rect2DActor->GetPositionCoordinate()
                                   ->GetComputedDisplayValue( this->CurrentRenderer );
      int* pos2 = this->Rect2DActor->GetPosition2Coordinate()
                                   ->GetComputedDisplayValue( this->CurrentRenderer );

    // Ignore event if the mouse is still outside the rectangle.
    if ( this->State == vtkRect2DWidget::Outside )
      {
      if ( X < pos1[0] || X > pos2[0] || Y < pos1[1] || Y > pos2[1] )
        {
        return;
        }
      // Mouse has moved inside the rectangle.
      this->State = vtkRect2DWidget::Inside;
      }
    else
      {
      // Mouse was previously inside the rectangle. Check if it has moved
      // outside.
      if ( X < pos1[0] || X > pos2[0] || Y < pos1[1] || Y > pos2[1] )
        {
        this->State = vtkRect2DWidget::Outside;
        this->Interactor->GetRenderWindow()->SetCurrentCursor( VTK_CURSOR_DEFAULT );
        return;
        }
      }

    // Mouse is inside the rectangle. Change the cursor to dragging unless it is
    // over an edge.

    this->SetCursor( this->ComputeStateBasedOnPosition( X, Y, pos1, pos2 ) );
    return;

/*    // If inside, set the cursor to the correct shape.
    if ( this->State == vtkRect2DWidget::Inside )
      {
      // if we have left then change cursor back to default
      if (X < pos1[0] || X > pos2[0] ||
          Y < pos1[1] || Y > pos2[1])
        {
        this->State = vtkRect2DWidget::Outside;
        this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_DEFAULT);
        return;
        }
      // adjust the cursor based on our position
      this->SetCursor(this->ComputeStateBasedOnPosition(X,Y,pos1,pos2));
      return;
      } */
    }
  
  // A drag operation was previously flagged.
  // Convert the display coordinates of the mouse to normalised viewport
  // coordinates to match the 2D actor.

  double XF = X;
  double YF = Y;
  this->CurrentRenderer->DisplayToNormalizedDisplay(XF,YF);
  this->CurrentRenderer->NormalizedDisplayToViewport(XF,YF);
  this->CurrentRenderer->ViewportToNormalizedViewport(XF,YF);
  
  // There are four parameters that can be adjusted.
  double *fpos1 = this->Rect2DActor->GetPositionCoordinate()->GetValue();
  double *fpos2 = this->Rect2DActor->GetPosition2Coordinate()->GetValue();
  double par1[2];
  double par2[2];
  par1[0] = fpos1[0];
  par1[1] = fpos1[1];
  par2[0] = fpos1[0] + fpos2[0];  
  par2[1] = fpos1[1] + fpos2[1];  
    
  // Based on the state, adjust the actor position and size.
  switch (this->State)
    {
    case vtkRect2DWidget::AdjustingP1:
      par1[0] = par1[0] + XF - this->StartPosition[0];
      par1[1] = par1[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::AdjustingP2:
      par2[0] = par2[0] + XF - this->StartPosition[0];
      par1[1] = par1[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::AdjustingP3:
      par2[0] = par2[0] + XF - this->StartPosition[0];
      par2[1] = par2[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::AdjustingP4:
      par1[0] = par1[0] + XF - this->StartPosition[0];
      par2[1] = par2[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::AdjustingE1:
      par1[0] = par1[0] + XF - this->StartPosition[0];
      break;
    case vtkRect2DWidget::AdjustingE2:
      par1[1] = par1[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::AdjustingE3:
      par2[0] = par2[0] + XF - this->StartPosition[0];
      break;
    case vtkRect2DWidget::AdjustingE4:
      par2[1] = par2[1] + YF - this->StartPosition[1];
      break;
    case vtkRect2DWidget::Moving:
      par1[0] = par1[0] + XF - this->StartPosition[0];
      par1[1] = par1[1] + YF - this->StartPosition[1];
      par2[0] = par2[0] + XF - this->StartPosition[0];
      par2[1] = par2[1] + YF - this->StartPosition[1];
      break;
    }
  
  // Push the change to the actor. Ensure  it doesn't shrink to nothing.
  if (par2[0] > par1[0] && par2[1] > par1[1])
    {
//     TRACE("vtkRect2DWidget::OnMouseMove() - par1 = (%e, %e), par2 = (%e, %e).\n",
//                                              par1[0], par1[1], par2[0], par2[1] );

    this->Rect2DActor->GetPositionCoordinate()->SetValue(par1[0],par1[1]);
    this->Rect2DActor->GetPosition2Coordinate()->
                       SetValue( par2[0] - par1[0], par2[1] - par1[1] );

    this->StartPosition[0] = XF;
    this->StartPosition[1] = YF;
    }
  
  // start a drag
  this->EventCallbackCommand->SetAbortFlag(1);
  this->InvokeEvent( vtkCommand::InteractionEvent, NULL );
  this->Interactor->Render();
}


void vtkRect2DWidget::OnLeftButtonUp()
{
  if (this->State == vtkRect2DWidget::Outside || this->LeftButtonDown == 0)
    {
    return;
    }

  // stop adjusting
  this->State = vtkRect2DWidget::Outside;
  this->EventCallbackCommand->SetAbortFlag(1);
  this->Interactor->GetRenderWindow()->SetCurrentCursor(VTK_CURSOR_DEFAULT);
  this->EndInteraction();
  this->InvokeEvent(vtkCommand::EndInteractionEvent,NULL);
  this->LeftButtonDown = 0;
}


void vtkRect2DWidget::OnRightButtonDown()
{

  // Ignore event if the mouse is outside the rect.
  if ( this->State == vtkRect2DWidget::Outside )
    {
    return;
    }

  if (this->HasObserver(vtkCommand::RightButtonPressEvent) ) 
    {
    this->EventCallbackCommand->SetAbortFlag(1);
    this->InvokeEvent(vtkCommand::RightButtonPressEvent,NULL);
    }
  RightButtonDown = 1;
}


void vtkRect2DWidget::OnRightButtonUp()
{
  if ( this->RightButtonDown == 0 )
    {
    return;
    }

  if (this->HasObserver(vtkCommand::RightButtonReleaseEvent)) 
    {
    this->EventCallbackCommand->SetAbortFlag(1);
    this->InvokeEvent(vtkCommand::RightButtonReleaseEvent,NULL);
    }
  this->RightButtonDown = 0;
}


void vtkRect2DWidget::GetRectNormalizedViewport( double rect[4] )
{
  GetRectNormalizedViewport( rect[0], rect[1], rect[2], rect[3] );
}


void vtkRect2DWidget::GetRectNormalizedViewport( double& xMin, double& yMin,
                                                 double& xMax, double& yMax )
{
  xMin = this->Rect2DActor->GetPosition()[0];
  yMin = this->Rect2DActor->GetPosition()[1];

  // Position2() is relative to Position() in placing the actor2D.
  xMax = xMin + this->Rect2DActor->GetPosition2()[0];
  yMax = yMin + this->Rect2DActor->GetPosition2()[1];
}


bool vtkRect2DWidget::GetRectWorld( double rect[4] )
{
  return GetRectWorld( rect[0], rect[1], rect[2], rect[3] );
}


bool vtkRect2DWidget::GetRectWorld( double& xMin, double& yMin,
                                    double& xMax, double& yMax )
{
  bool copied = false;

  // Normally, the renderer is determined from the last event position when
  // the widget is enabled (this->Interactor->FindPokedRenderer()) and is
  // nulled when the widget is disabled.

  if ( this->Enabled )
  {
     double* pos1 = this->Rect2DActor->GetPositionCoordinate()
                                     ->GetComputedWorldValue( this->CurrentRenderer );
     xMin = pos1[0];
     yMin = pos1[1];

     // Although Position2() is relative to Position() in placing the actor2D,
     // the computed value is relative to the origin.

     double* pos2 = this->Rect2DActor->GetPosition2Coordinate()
                                     ->GetComputedWorldValue( this->CurrentRenderer );
     xMax = pos2[0];
     yMax = pos2[1];
  }
  return copied;
}


bool vtkRect2DWidget::GetRectViewport( int rect[4] )
{
  return GetRectViewport( rect[0], rect[1], rect[2], rect[3] );
}


bool vtkRect2DWidget::GetRectViewport( int& xMin, int& yMin, int& xMax, int& yMax )
{
  bool copied = false;
  if ( this->Enabled )
  {
     int* pos1 = this->Rect2DActor->GetPositionCoordinate()
                                  ->GetComputedViewportValue( this->CurrentRenderer );
     xMin = pos1[0];
     yMin = pos1[1];

     int* pos2 = this->Rect2DActor->GetPosition2Coordinate()
                                  ->GetComputedViewportValue( this->CurrentRenderer );
     xMax = pos2[0];
     yMax = pos2[1];
  }
  return copied;
}


bool vtkRect2DWidget::GetRectDisplay( int rect[4] )
{
  return GetRectDisplay( rect[0], rect[1], rect[2], rect[3] );
}


bool vtkRect2DWidget::GetRectDisplay( int& xMin, int& yMin, int& xMax, int& yMax )
{
  bool copied = false;
  if ( this->Enabled )
  {
     int* pos1 = this->Rect2DActor->GetPositionCoordinate()
                                  ->GetComputedDisplayValue( this->CurrentRenderer );
     xMin = pos1[0];
     yMin = pos1[1];

     int* pos2 = this->Rect2DActor->GetPosition2Coordinate()
                                  ->GetComputedDisplayValue( this->CurrentRenderer );
     xMax = pos2[0];
     yMax = pos2[1];
  }
  return copied;
}


void vtkRect2DWidget::PrintSelf( ostream& os, vtkIndent indent )
{
  this->Superclass::PrintSelf(os,indent);
}




