/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkRect2DWidget.h,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkRect2DWidget - 2D widget for rectangular selection
// .SECTION Description
// This class provides interactive manipulation of the position and size of
// vtkRect2DActor. It listens to Left mouse events and mouse movement. It
// also listens to Right mouse events and notifies any observers of Right
// mouse events on this object when they occur.
//
// The cursor shape changes to a resize cursor if it is over the border of
// the rectangle or a move cursor when inside the rectangle .

// .SECTION See Also
// vtkInteractorObserver

#ifndef __vtkRect2DWidget_h
#define __vtkRect2DWidget_h

class vtkProperty2D;

#include "vtkInteractorObserver.h"
#include "vtkRect2DActor.h"

class VTK_HYBRID_EXPORT vtkRect2DWidget : public vtkInteractorObserver
{
public:
  // Description:
  // Instantiate the widget.
  static vtkRect2DWidget *New();

  vtkTypeRevisionMacro(vtkRect2DWidget,vtkInteractorObserver);
  void PrintSelf(ostream& os, vtkIndent indent);

  // Description:
  // Define the initial position of the widget in normalised viewport
  // coordinates. The default position and size is the same as vtkActor2D
  // ( 0.0, 0.0, 0.5, 0.5 ).
  virtual void PlaceRect( const double xMin, const double yMin,
                          const double xMax, const double yMax );

  // Description:
  // Enable/disable the widget. The rect actor will not be visible unless
  // enabled.
  virtual void SetEnabled(int);

  // Description:
  // Set/Get the 2D rect actor.
  vtkGetObjectMacro( Rect2DActor, vtkRect2DActor );

  // Description:
  // Get the current position of the rect actor in normalised viewport coordinates.
  virtual void GetRectNormalizedViewport( double rect[4] );
  virtual void GetRectNormalizedViewport( double& xMin, double& yMin,
                                          double& xMax, double& yMax );

  // Description:
  // Get the current position of the rect actor in world, viewport or display
  // coordinates. If the widget is not enabled, the return value is false and
  // the coordinates are not set.
  virtual bool GetRectWorld( double rect[4] );
  virtual bool GetRectWorld( double& xMin, double& yMin, double& xMax, double& yMax );
  virtual bool GetRectViewport( int rect[4] );
  virtual bool GetRectViewport( int& xMin, int& yMin, int& xMax, int& yMax );
  virtual bool GetRectDisplay( int rect[4] );
  virtual bool GetRectDisplay( int& xMin, int& yMin, int& xMax, int& yMax );

protected:
  vtkRect2DWidget();
  ~vtkRect2DWidget();

  // Handles the mouse events.
  static void ProcessEvents( vtkObject* object, unsigned long event,
                                   void* clientdata, void* calldata );

  // ProcessEvents() dispatches to these methods.
  void OnLeftButtonDown();
  void OnLeftButtonUp();
  void OnRightButtonDown();
  void OnRightButtonUp();
  void OnMouseMove();

  // Compute relative movements
  double StartPosition[2];
  
//BTX - manage the state of the widget
  int State;
  // use this to track whether left/right button was pressed to gate
  // action on button up event.
  int LeftButtonDown;
  int RightButtonDown;
  enum WidgetState
  {
    Moving=0,
    AdjustingP1,
    AdjustingP2,
    AdjustingP3,
    AdjustingP4,
    AdjustingE1,
    AdjustingE2,
    AdjustingE3,
    AdjustingE4,
    Inside,
    Outside
  };
//ETX

  // Determine the WidgetState based on mouse location, edge1 p1, etc.
  int ComputeStateBasedOnPosition( int X, int Y, int* pos1, int* pos2 );

  // Set the cursor shape based on the State argument.
  void SetCursor(int State);

  // The actor that is changed by the mouse events.
  vtkRect2DActor* Rect2DActor;   

  vtkProperty2D*       BorderProperty;
  vtkProperty2D*       BackgroundProperty;

private:
  vtkRect2DWidget(const vtkRect2DWidget&);  //Not implemented
  void operator=(const vtkRect2DWidget&);  //Not implemented
};

#endif
