/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkRect2DActor.cxx,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "stdafx.h"
#include "vtkRect2DActor.h"

#include "vtkCellArray.h"
#include "vtkObjectFactory.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper2D.h"
#include "vtkProperty2D.h"
#include "vtkViewport.h"
#include "vtkWindow.h"

#include "vtkCommand.h"

vtkCxxRevisionMacro(vtkRect2DActor, "$Revision: 1.00 $");
vtkStandardNewMacro(vtkRect2DActor);

vtkCxxSetObjectMacro(vtkRect2DActor, BorderProperty, vtkProperty2D);
vtkCxxSetObjectMacro(vtkRect2DActor, BackgroundProperty, vtkProperty2D);

vtkRect2DActor::vtkRect2DActor()
{
  this->ShowBorder = 1; 
  this->ShowBackground = 0; 

  // The default coordinate systems for positioning 2D actors are VIEWPORT
  // coordinates for SetPosition() and NORMALISED VIEWPORT relative to Position
  // for SetPosition2(). Change SetPosition() to use NORMALISED VIEWPORT.
  // The initial position and size of the rectangle is same as the vtkActor2D
  // defaults (0, 0) -> (0.5, 0.5).

  this->PositionCoordinate->SetCoordinateSystemToNormalizedViewport();

  // Create the points for the border. By default, vtkPolyDataMapper2D,
  // transforms 3D data into 2D data by ignoring the z-coordinate of the 3D
  // points in vtkPolyData, and taking the (x,y) values as VIEWPORT values 
  // (pixel centres relative to the lower-left corner of the viewport in the
  // render window). Exploit this default behaviour by defining the polydata in
  // VIEWPORT coordinates. Note that the actual positions are computed at the
  // point of rendering.

  vtkPoints* borderPts = vtkPoints::New( VTK_DOUBLE ); // stores 1/2 integer positions
  borderPts->SetNumberOfPoints( 4 );

  // Create polydata to represent the border of the rectangle. Separate polydata
  // is required for the border and background because vtkPolyDataMapper2D does
  // not have a wireframe surface representation.

  this->BorderPolyData = vtkPolyData::New();
  this->BorderPolyData->SetPoints( borderPts );
  borderPts->Delete();

  // Create a polyline representing the border of the rectangle by first
  // specifying the number of points and then adding the zero-based offset
  // to each point Id in the associated point list.

  vtkCellArray* border = vtkCellArray::New();
  border->InsertNextCell( 5 );
  border->InsertCellPoint( 0 );
  border->InsertCellPoint( 1 );
  border->InsertCellPoint( 2 );
  border->InsertCellPoint( 3 );
  border->InsertCellPoint( 0 );

  this->BorderPolyData->SetLines( border );
  border->Delete();

  this->BorderMapper = vtkPolyDataMapper2D::New();
  this->BorderMapper->SetInput( this->BorderPolyData );
  this->BorderActor = vtkActor2D::New();
  this->BorderActor->SetMapper( this->BorderMapper );
  this->BorderActor->SetVisibility( this->ShowBorder );

  // Set the coordinates of the compositing actors relative to the parent actor.
  this->BorderActor->GetPositionCoordinate()->
                     SetReferenceCoordinate( this->PositionCoordinate );

  // Create the default border properties. Note that vtkActor2D does not create
  // default properties on construction.

  this->BorderProperty = vtkProperty2D::New();
  this->BorderProperty->SetLineWidth( 1 );
  this->BorderProperty->SetColor( 0.0, 0.0, 0.0 );
  this->BorderProperty->SetOpacity( 1.0 );
  this->BorderActor->SetProperty( this->BorderProperty );

  // Create the points for the background. These must be deflated to fit inside
  // the border otherwise the background obscures the border. 
  // vtkPolyDataMapper2D cannot resolve coincident topology.

  vtkPoints* backgroundPts = vtkPoints::New( VTK_DOUBLE );
  backgroundPts->SetNumberOfPoints( 4 );

  // Create polydata to represent the background of the rectangle.
  this->BackgroundPolyData = vtkPolyData::New();
  this->BackgroundPolyData->SetPoints( backgroundPts );
  backgroundPts->Delete();

  // Create a single quadrilateral for the background. 
  vtkCellArray* background = vtkCellArray::New();
  background->InsertNextCell( 4 );
  background->InsertCellPoint( 0 );
  background->InsertCellPoint( 1 );
  background->InsertCellPoint( 2 );
  background->InsertCellPoint( 3 );

  this->BackgroundPolyData->SetPolys( background );
  background->Delete();

  this->BackgroundMapper = vtkPolyDataMapper2D::New();
  this->BackgroundMapper->SetInput( this->BackgroundPolyData );
  this->BackgroundActor = vtkActor2D::New();
  this->BackgroundActor->SetMapper( this->BackgroundMapper );
  this->BackgroundActor->SetVisibility( this->ShowBackground );

  // Set the coordinates of the compositing actors relative to the parent actor.
  this->BackgroundActor->GetPositionCoordinate()->
                         SetReferenceCoordinate( this->PositionCoordinate );

  // Create the default background properties.
  this->BackgroundProperty = vtkProperty2D::New();
  this->BackgroundProperty->SetColor( 1.0, 1.0, 1.0 );
  this->BackgroundProperty->SetOpacity( 0.3 );
  this->BackgroundActor->SetProperty( this->BackgroundProperty );

  // Record the last position of the actor2D in VIEWPORT coordinates (pixel
  // centres relative to the lower-left corner of the viewport in the render
  // window) rounded to the nearest integer.

  this->LastPosition[0]  = 0;
  this->LastPosition[1]  = 0;
  this->LastPosition2[0] = 0;
  this->LastPosition2[1] = 0;
}


vtkRect2DActor::~vtkRect2DActor()
{
  this->BorderProperty->Delete();
  this->BorderActor->Delete();
  this->BorderMapper->Delete();
  this->BorderPolyData->Delete();

  this->BackgroundProperty->Delete();
  this->BackgroundActor->Delete();
  this->BackgroundMapper->Delete();
  this->BackgroundPolyData->Delete();
}


void vtkRect2DActor::SetShowBorder( const int show )
{
   if ( show != this->ShowBorder )
   {
      this->SetVisibility( show );
      this->ShowBorder = show;
      this->Modified();
   }
}


void vtkRect2DActor::SetShowBackground( const int show )
{
   if ( show != this->ShowBackground )
   {
      this->SetVisibility( show );
      this->ShowBackground = show;
      this->Modified();
   }
}


void vtkRect2DActor::ReleaseGraphicsResources( vtkWindow* win )
//-----------------------------------------------------------------------------
//
// Release any graphics resources that are being consumed by this actor. The
// parameter window could be used to determine which graphic resources to
// release.
//
//-----------------------------------------------------------------------------
{
  this->BorderActor->ReleaseGraphicsResources( win );
  this->BackgroundActor->ReleaseGraphicsResources( win );
}


int vtkRect2DActor::RenderOverlay( vtkViewport* viewport )
{
  int renderedSomething = 0;
  renderedSomething += this->BorderActor->RenderOverlay( viewport );
  renderedSomething += this->BackgroundActor->RenderOverlay( viewport );

  return ( renderedSomething > 0 ) ? 1 : 0;
}


int vtkRect2DActor::RenderOpaqueGeometry( vtkViewport *viewport )
{
  int renderedSomething = 0;
  
//  TRACE( "viewport->GetMTime() = %d.\n", viewport->GetMTime() );
//  TRACE( "viewport->GetVTKWindow()->GetMTime() = %d.\n", viewport->GetVTKWindow()->GetMTime() );

  // If the viewport has changed, the border and background polydata points may
  // or may not need updating. It depends if the projected coordinates have
  // changed. Note that the viewport and window are modified by mouse activity
  // in the render window.

  int positionsHaveChanged = 0;
  int* position  = NULL;
  int* position2 = NULL;
  if ( viewport->GetMTime() > this->BuildTime || 
      ( viewport->GetVTKWindow() && viewport->GetVTKWindow()->GetMTime() > this->BuildTime ) )
    {
    position  = this->PositionCoordinate->GetComputedViewportValue( viewport );
    position2 = this->Position2Coordinate->GetComputedViewportValue( viewport );
    if ( this->LastPosition[0]  != position[0]  || this->LastPosition[1]  != position[1] ||
         this->LastPosition2[0] != position2[0] || this->LastPosition2[1] != position2[1] )
      {
      positionsHaveChanged = 1;
      }
    }
  
  // The polydata points must also be updated if the border line thickness has
  // changed.

  if ( positionsHaveChanged || this->GetMTime() > this->BuildTime )
    {
    vtkDebugMacro( << "Updating the border and background polydata points" );

    SizeRect2D( viewport );

    // Save the current actor position to check for position changes on the next
    // render.

    if ( NULL == position )
      position  = this->PositionCoordinate->GetComputedViewportValue( viewport );
    if ( NULL == position2 )
      position2 = this->Position2Coordinate->GetComputedViewportValue( viewport );

    this->LastPosition[0]  = position[0];
    this->LastPosition[1]  = position[1];
    this->LastPosition2[0] = position2[0];
    this->LastPosition2[1] = position2[1];

    this->BuildTime.Modified();
    }

  renderedSomething += this->BorderActor->RenderOpaqueGeometry( viewport );
  renderedSomething += this->BackgroundActor->RenderOpaqueGeometry( viewport );

  return ( renderedSomething > 0 ) ? 1 : 0;
}


void vtkRect2DActor::SizeRect2D( vtkViewport* viewport )
// -----------------------------------------------------------------------------
//
// Size the border and background polydata to precisely coincide with the new
// position of the actor. To use the default behaviour of vtkPolyDataMapper2D,
// the polydata points must be presented in VIEWPORT values (x,y pixel centres
// relative to the lower-left corner of the viewport in the render window). 
//
// -----------------------------------------------------------------------------
{
  // Set the bottom, left point of the border.
  double X0 = 0.5;
  double Y0 = 0.5;

  // Set the top, right point of the border.
//  double* pos1 = this->BorderActor->GetPositionCoordinate()->GetComputedDoubleViewportValue( this->CurrentRenderer );
//  double* pos2 = this->BorderActor->GetPosition2Coordinate()->GetComputedDoubleViewportValue( this->CurrentRenderer );
  int* pos1 = this->GetPositionCoordinate()->GetComputedViewportValue( viewport );
  int* pos2 = this->GetPosition2Coordinate()->GetComputedViewportValue( viewport );
  double X1 = pos2[0] - pos1[0];
  double Y1 = pos2[1] - pos1[1];

  vtkPoints* pts = this->BorderPolyData->GetPoints();
  pts->SetPoint( 0, X0,      Y0,      0 );
  pts->SetPoint( 1, X0 + X1, Y0,      0 );
  pts->SetPoint( 2, X0 + X1, Y0 + Y1, 0 );
  pts->SetPoint( 3, X0,      Y0 + Y1, 0 );

  // Position the background. This must not overlap the border. Note that
  // offsets for the left and top edges may be different from the bottom and
  // right edges.
   
  int borderWidth = this->BorderActor->GetProperty()->GetLineWidth();
  double offLT = static_cast<int>( ( borderWidth + 1 )/2);
  double offRB = static_cast<int>( ( borderWidth     )/2);

  pts = this->BackgroundPolyData->GetPoints();
  pts->SetPoint( 0, X0      + offLT, Y0      + offRB, 0 );
  pts->SetPoint( 1, X0 + X1 - offRB, Y0      + offRB, 0 );
  pts->SetPoint( 2, X0 + X1 - offRB, Y0 + Y1 - offLT, 0 );
  pts->SetPoint( 3, X0      + offLT, Y0 + Y1 - offLT, 0 );

  // This is here to notify the widget that the actor position/size has changed.
  // This could be due to an interaction with the widget or resizing of the
  // render window. 
//  InvokeEvent( vtkCommand::UserEvent + 99, NULL );
  InvokeEvent( vtkCommand::ModifiedEvent, NULL );
}


void vtkRect2DActor::ShallowCopy( vtkProp* prop )
{
  vtkRect2DActor* a = vtkRect2DActor::SafeDownCast( prop );
  if ( a != NULL )
    {
    this->SetShowBorder( a->GetShowBorder() ); 
    this->SetShowBackground( a->GetShowBackground() ); 
    this->SetBorderProperty( a->GetBorderProperty() );
    this->SetBackgroundProperty( a->GetBackgroundProperty() );
    this->SetPosition2 (a->GetPosition2() );
    this->GetPositionCoordinate()->SetCoordinateSystem(
       a->GetPositionCoordinate()->GetCoordinateSystem() );    
    this->GetPositionCoordinate()->SetValue(
       a->GetPositionCoordinate()->GetValue() );
    this->GetPosition2Coordinate()->SetCoordinateSystem(
       a->GetPosition2Coordinate()->GetCoordinateSystem() );    
    this->GetPosition2Coordinate()->SetValue(
       a->GetPosition2Coordinate()->GetValue() );
    }

  // Now do superclass.
  this->vtkActor2D::ShallowCopy( prop );
}


void vtkRect2DActor::PrintSelf( ostream& os, vtkIndent indent )
{
  this->Superclass::PrintSelf( os, indent );

  os << indent << "ShowBorder: " << this->ShowBorder << "\n";
  os << indent << "ShowBackground: " << this->ShowBackground << "\n";
}


